@@ -1,9 +1,9 @@
 package Net::DNS::Nameserver;
 
 #
-# $Id: Nameserver.pm 1526 2017-01-16 09:17:54Z willem $
+# $Id: Nameserver.pm 1605 2017-11-27 11:37:40Z willem $
 #
-our $VERSION = (qw$LastChangedRevision: 1526 $)[1];
+our $VERSION = (qw$LastChangedRevision: 1605 $)[1];
 
 
 =head1 NAME
@@ -14,23 +14,32 @@ Net::DNS::Nameserver - DNS server class
 
     use Net::DNS::Nameserver;
 
-    $nameserver = new Net::DNS::Nameserver(
-	LocalAddr	 => ['::1' , '127.0.0.1' ],
-	LocalPort	 => "5353",
-	ReplyHandler => \&reply_handler,
-	Verbose		 => 1,
-	Truncate	 => 0
+    my $nameserver = new Net::DNS::Nameserver(
+	LocalAddr	=> ['::1' , '127.0.0.1'],
+	ZoneFile	=> "filename"
+	);
+
+    my $nameserver = new Net::DNS::Nameserver(
+	LocalAddr	=> '10.1.2.3',
+	LocalPort	=> 5353,
+	ReplyHandler	=> \&reply_handler
     );
 
 
 =head1 DESCRIPTION
 
-Instances of the C<Net::DNS::Nameserver> class represent DNS server
-objects.  See L</EXAMPLE> for an example.
+Net::DNS::Nameserver offers a simple mechanism for instantiation of
+customised DNS server objects intended to provide test responses to
+queries emanating from a client resolver.
+
+It is not, nor will it ever be, a general-purpose DNS nameserver
+implementation.
+
+See L</EXAMPLE> for an example.
 
 =cut
 
-use constant USE_SOCKET_IP => defined eval 'use Socket 1.98; use IO::Socket::IP 0.32; 1;';
+use constant USE_SOCKET_IP => defined eval 'use Socket 1.97; use IO::Socket::IP 0.32; 1;';
 
 use constant USE_SOCKET_INET => defined eval 'require IO::Socket::INET';
 
@@ -42,15 +51,16 @@ use constant IPv6 => USE_SOCKET_IP || USE_SOCKET_INET6;
 use strict;
 use warnings;
 use integer;
-use Carp qw(cluck);
+use Carp;
 use Net::DNS;
+use Net::DNS::ZoneFile;
 
 use IO::Socket;
 use IO::Select;
 
 use constant FORCE_IPv4 => 0;
 
-use constant DEFAULT_ADDR => 0;
+use constant DEFAULT_ADDR => qw(::1 127.0.0.1);
 use constant DEFAULT_PORT => 53;
 
 use constant STATE_ACCEPTED   => 1;
@@ -72,17 +82,13 @@ sub new {
 			$self{ReplyHandler} = sub { $handler->( $self, @_ ); };
 		}
 	}
-	unless ( ref $self{ReplyHandler} eq "CODE" ) {
-		cluck "No reply handler!";
-		return undef;
-	}
+	croak 'No reply handler!' unless ref( $self{ReplyHandler} ) eq "CODE";
+
+	$self->ReadZoneFile( $self{ZoneFile} ) if exists $self{ZoneFile};
 
 	# local server addresses must also be accepted by a resolver
-	my @LocalAddr =
-			ref $self{LocalAddr}
-			? @{$self{LocalAddr}}
-			: ( $self{LocalAddr} || DEFAULT_ADDR );
-	my $resolver = new Net::DNS::Resolver( nameservers => [@LocalAddr] );
+	my $LocalAddr = $self{LocalAddr} || [DEFAULT_ADDR];
+	my $resolver = new Net::DNS::Resolver( nameservers => $LocalAddr );
 	$resolver->force_v4(1) if FORCE_IPv4;
 	my @localaddresses = $resolver->nameservers;
 
@@ -116,7 +122,7 @@ sub new {
 			push @sock_tcp, $sock_tcp;
 			print "done.\n" if $self{Verbose};
 		} else {
-			cluck "Couldn't create TCP socket: $!";
+			carp "Couldn't create TCP socket: $!";
 		}
 
 		#--------------------------------------------------------------------------
@@ -135,7 +141,7 @@ sub new {
 			push @sock_udp, $sock_udp;
 			print "done.\n" if $self{Verbose};
 		} else {
-			cluck "Couldn't create UDP socket: $!";
+			carp "Couldn't create UDP socket: $!";
 		}
 
 	}
@@ -158,6 +164,67 @@ sub new {
 	return $self;
 }
 
+
+#------------------------------------------------------------------------------
+# ReadZoneFile - Read zone file used by default reply handler
+#------------------------------------------------------------------------------
+
+sub ReadZoneFile {
+	my ( $self, $file ) = @_;
+	my $zonefile = new Net::DNS::ZoneFile($file);
+
+	my $RRhash = $self->{RRhash} = {};
+	my $RRlist = [];
+	while ( my $rr = $zonefile->read ) {
+		my ($leaf) = $rr->{owner}->label;
+		push @{$RRhash->{lc $leaf}}, $rr;
+
+		# Warning: Nasty trick abusing SOA to reference zone RR list
+		if ( $rr->type eq 'SOA' ) { $RRlist = $rr->{RRlist} = [] }
+		else			  { push @$RRlist, $rr }
+	}
+}
+
+
+#------------------------------------------------------------------------------
+# ReplyHandler - Default reply handler serving RRs from zone file
+#------------------------------------------------------------------------------
+
+sub ReplyHandler {
+	my ( $self, $qname, $qclass, $qtype, $peerhost, $query, $conn ) = @_;
+	my $opcode = $query->header->opcode;
+	my $rcode  = 'NOERROR';
+	my @ans;
+
+	my $lcase = lc $qname;					# assume $qclass always 'IN'
+	my ( $leaf, @tail ) = split /\./, $lcase;
+	my $RRhash = $self->{RRhash};
+	my $RRlist = $RRhash->{$leaf} || [];			# hash, then linear search
+	my @match  = grep lc( $_->owner ) eq $lcase, @$RRlist;
+
+	if ( $qtype eq 'AXFR' ) {
+		my ($soa) = grep $_->type eq 'SOA', @match;
+		if ($soa) { push @ans, $soa, @{$soa->{RRlist}}, $soa }
+		else	  { $rcode = 'NOTAUTH' }
+
+	} else {
+		unless ( scalar(@match) ) {
+			my $wildcard = join '.', '*', @tail;
+			my $wildlist = $RRhash->{'*'} || [];
+			foreach ( grep lc( $_->owner ) eq $wildcard, @$wildlist ) {
+				my $clone = bless {%$_}, ref($_);
+				$clone->owner($qname);
+				push @match, $clone;
+			}
+			$rcode = 'NXDOMAIN' unless @match;
+		}
+		@ans = grep $_->type eq $qtype, @match;
+	}
+
+	return ( $rcode, \@ans, [], [], {aa => 1}, {} );
+}
+
+
 #------------------------------------------------------------------------------
 # inet_new - Calls the constructor in the correct module for making sockets.
 #------------------------------------------------------------------------------
@@ -196,6 +263,7 @@ sub make_reply {
 	my $reply  = $query->reply();
 	my $header = $reply->header;
 	my $headermask;
+	my $optionmask;
 
 	my $opcode  = $query->header->opcode;
 	my $qdcount = $query->header->qdcount;
@@ -214,18 +282,18 @@ sub make_reply {
 		my $qclass = $qr->qclass;
 
 		my $id = $query->header->id;
-		print "query $id : $qname $qclass $qtype - " if $self->{Verbose};
+		print "query $id : $qname $qclass $qtype\n" if $self->{Verbose};
 
 		my ( $rcode, $ans, $auth, $add );
 		my @arglist = ( $qname, $qclass, $qtype, $peerhost, $query, $conn );
 
 		if ( $opcode eq "QUERY" ) {
-			( $rcode, $ans, $auth, $add, $headermask ) =
+			( $rcode, $ans, $auth, $add, $headermask, $optionmask ) =
 					&{$self->{ReplyHandler}}(@arglist);
 
 		} elsif ( $opcode eq "NOTIFY" ) {		#RFC1996
 			if ( ref $self->{NotifyHandler} eq "CODE" ) {
-				( $rcode, $ans, $auth, $add, $headermask ) =
+				( $rcode, $ans, $auth, $add, $headermask, $optionmask ) =
 						&{$self->{NotifyHandler}}(@arglist);
 			} else {
 				$rcode = "NOTIMP";
@@ -233,7 +301,7 @@ sub make_reply {
 
 		} elsif ( $opcode eq "UPDATE" ) {		#RFC2136
 			if ( ref $self->{UpdateHandler} eq "CODE" ) {
-				( $rcode, $ans, $auth, $add, $headermask ) =
+				( $rcode, $ans, $auth, $add, $headermask, $optionmask ) =
 						&{$self->{UpdateHandler}}(@arglist);
 			} else {
 				$rcode = "NOTIMP";
@@ -260,7 +328,11 @@ sub make_reply {
 		$header->$key($value);
 	}
 
-	$header->print if $self->{Verbose} && defined $headermask;
+	while ( my ( $option, $value ) = each %{$optionmask || {}} ) {
+		$reply->edns->option( $option, $value );
+	}
+
+	$header->print if $self->{Verbose} && ( $headermask || $optionmask );
 
 	return $reply;
 }
@@ -329,9 +401,7 @@ sub tcp_connection {
 				return;				# Still not 2 octets ready
 			}
 			my $msglen = unpack( "n", $1 );
-			print "Removed 2 octets from the input buffer from $peer.\n"
-					. "$peer said his query contains $msglen octets.\n"
-					if $self->{Verbose};
+			print "$peer said his query contains $msglen octets\n" if $self->{Verbose};
 			$self->{_tcp}{$sock}{state}	  = STATE_GOT_LENGTH;
 			$self->{_tcp}{$sock}{querylength} = $msglen;
 		}
@@ -364,10 +434,11 @@ sub tcp_connection {
 				delete $self->{_tcp}{$sock};
 				return;
 			}
-			my $reply_data = $reply->data;
-			my $len	       = length $reply_data;
-			$self->{_tcp}{$sock}{outbuffer} = pack( "n", $len ) . $reply_data;
-			print "Queued ", length $self->{_tcp}{$sock}{outbuffer}, " octets to $peer\n"
+			my $reply_data = $reply->data(65535);	# limit to one TCP envelope
+			warn "multi-packet TCP response not implemented" if $reply->header->tc;
+			my $len = length $reply_data;
+			$self->{_tcp}{$sock}{outbuffer} = pack( 'n a*', $len, $reply_data );
+			print "Queued TCP response (2 + $len octets) to $peer\n"
 					if $self->{Verbose};
 
 			# We are done.
@@ -396,7 +467,7 @@ sub udp_connection {
 
 	print "UDP connection from $peerhost:$peerport to $sockhost\n" if $self->{Verbose};
 
-	my $query = new Net::DNS::Packet( \$buf, $self->{Verbose} );
+	my $query = new Net::DNS::Packet( \$buf );
 	if ( my $err = $@ ) {
 		print "Error decoding query packet: $err\n" if $self->{Verbose};
 		undef $query;					# force FORMERR reply
@@ -412,7 +483,7 @@ sub udp_connection {
 	my $max_len = ( $query && $self->{Truncate} ) ? $query->edns->size : undef;
 	if ( $self->{Verbose} ) {
 		local $| = 1;
-		print "Maximum UDP size advertised by $peerhost:$peerport: $max_len\n" if $max_len;
+		print "Maximum UDP size advertised by $peerhost#$peerport: $max_len\n" if $max_len;
 		print "Writing response - ";
 		print $sock->send( $reply->data($max_len) ) ? "done" : "failed: $!", "\n";
 
@@ -551,53 +622,48 @@ __END__
 
 =head2 new
 
-    my $ns = new Net::DNS::Nameserver(
-	LocalAddr	=> "10.1.2.3",
-	LocalPort	=> "5353",
-	ReplyHandler	=> \&reply_handler,
-	Verbose		=> 1
+    $nameserver = new Net::DNS::Nameserver(
+	LocalAddr	=> ['::1' , '127.0.0.1'],
+	ZoneFile	=> "filename"
 	);
 
-
-
-    my $ns = new Net::DNS::Nameserver(
-	LocalAddr	=> ['::1' , '127.0.0.1' ],
-	LocalPort	=> "5353",
+    $nameserver = new Net::DNS::Nameserver(
+	LocalAddr	=> '10.1.2.3',
+	LocalPort	=> 5353,
 	ReplyHandler	=> \&reply_handler,
 	Verbose		=> 1,
 	Truncate	=> 0
-	);
-
+    );
 
 Returns a Net::DNS::Nameserver object, or undef if the object
 could not be created.
 
-Attributes are:
+Each instance is configured using the following optional arguments:
 
-    LocalAddr		IP address on which to listen.	Defaults to INADDR_ANY.
-    LocalPort		Port on which to listen.	Defaults to 53.
-    ReplyHandler	Reference to reply-handling
-			subroutine			Required.
+    LocalAddr		IP address on which to listen	Defaults to loopback address
+    LocalPort		Port on which to listen		Defaults to 53
+    ZoneFile		Name of file containing RRs
+			accessed using the default
+			reply-handling subroutine
+    ReplyHandler	Reference to customised
+			reply-handling subroutine
     NotifyHandler	Reference to reply-handling
 			subroutine for queries with
 			opcode NOTIFY (RFC1996)
     UpdateHandler	Reference to reply-handling
 			subroutine for queries with
 			opcode UPDATE (RFC2136)
-    Verbose		Print info about received
-			queries.			Defaults to 0 (off).
+    Verbose		Report internal activity	Defaults to 0 (off)
     Truncate		Truncates UDP packets that
 			are too big for the reply	Defaults to 1 (on)
     IdleTimeout		TCP clients are disconnected
 			if they are idle longer than
-			this duration.			Defaults to 120 (secs)
+			this duration			Defaults to 120 (secs)
 
 The LocalAddr attribute may alternatively be specified as a list of IP
 addresses to listen to.
-
-If IO::Socket::INET6 and Socket6 are available on the system you can
-also list IPv6 addresses and the default is '0' (listen on all interfaces on
-IPv6 and IPv4);
+If the IO::Socket::IP library package is available on the system
+this may also include IPv6 addresses.
 
 
 The ReplyHandler subroutine is passed the query name, query class,
@@ -619,6 +685,9 @@ hashref with the settings for the C<aa>, C<ra>, and C<ad>
 header bits. The argument is of the form
 C<< { ad => 1, aa => 0, ra => 1 } >>.
 
+EDNS options may be specified in a similar manner using optionmask
+C<< { $optioncode => $value, $optionname => $value } >>.
+
 
 See RFC 1035 and the IANA dns-parameters file for more information:
 
@@ -700,34 +769,41 @@ additional filtering on its basis may be applied.
     use Net::DNS::Nameserver;
 
     sub reply_handler {
-	my ($qname, $qclass, $qtype, $peerhost,$query,$conn) = @_;
-	my ($rcode, @ans, @auth, @add);
+	my ( $qname, $qclass, $qtype, $peerhost, $query, $conn ) = @_;
+	my ( $rcode, @ans, @auth, @add );
 
-	print "Received query from $peerhost to ". $conn->{sockhost}. "\n";
+	print "Received query from $peerhost to " . $conn->{sockhost} . "\n";
 	$query->print;
 
-	if ($qtype eq "A" && $qname eq "foo.example.com" ) {
-		my ($ttl, $rdata) = (3600, "10.1.2.3");
+	if ( $qtype eq "A" && $qname eq "foo.example.com" ) {
+		my ( $ttl, $rdata ) = ( 3600, "10.1.2.3" );
 		my $rr = new Net::DNS::RR("$qname $ttl $qclass $qtype $rdata");
 		push @ans, $rr;
 		$rcode = "NOERROR";
-	}elsif( $qname eq "foo.example.com" ) {
+	} elsif ( $qname eq "foo.example.com" ) {
 		$rcode = "NOERROR";
 
-	}else{
+	} else {
 		$rcode = "NXDOMAIN";
 	}
 
-	# mark the answer as authoritative (by setting the 'aa' flag
-	return ($rcode, \@ans, \@auth, \@add, { aa => 1 });
+	# mark the answer as authoritative (by setting the 'aa' flag)
+	my $headermask = {aa => 1};
+
+	# specify EDNS options	{ option => value }
+	my $optionmask = {};
+
+	return ( $rcode, \@ans, \@auth, \@add, $headermask, $optionmask );
     }
 
+
     my $ns = new Net::DNS::Nameserver(
 	LocalPort    => 5353,
 	ReplyHandler => \&reply_handler,
 	Verbose	     => 1
 	) || die "couldn't create nameserver object\n";
 
+
     $ns->main_loop;
 
 
@@ -759,6 +835,8 @@ Portions Copyright (c)2005 Robert Martin-Legene.
 
 Portions Copyright (c)2005-2009 O.M, Kolkman, RIPE NCC.
 
+Portions Copyright (c)2017 Dick Franks.
+
 All rights reserved.
 
 
